@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -27,6 +27,7 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+#include <errno.h>
 
 #include <glib.h>
 
@@ -57,6 +58,9 @@ static const char *cpin_prefix[] = { "+CPIN:", NULL };
 static const char *clck_prefix[] = { "+CLCK:", NULL };
 static const char *huawei_cpin_prefix[] = { "^CPIN:", NULL };
 static const char *xpincnt_prefix[] = { "+XPINCNT:", NULL };
+static const char *zpinpuk_prefix[] = { "+ZPINPUK:", NULL };
+static const char *cpinr_prefixes[] = { "+CPINR:", "+CPINRE:", NULL };
+static const char *epin_prefix[] = { "*EPIN:", NULL };
 static const char *none_prefix[] = { NULL };
 
 static void at_crsm_info_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -142,13 +146,14 @@ static void at_sim_read_info(struct ofono_sim *sim, int fileid,
 	}
 
 	cbd = cb_data_new(cb, data);
-	if (cbd == NULL)
-		goto error;
 
 	snprintf(buf, sizeof(buf), "AT+CRSM=192,%i", fileid);
 
 	switch (sd->vendor) {
+	case OFONO_VENDOR_ZTE:
 	case OFONO_VENDOR_HUAWEI:
+	case OFONO_VENDOR_SIERRA:
+	case OFONO_VENDOR_SPEEDUP:
 	case OFONO_VENDOR_QUALCOMM_MSM:
 		strcat(buf, ",0,0,255"); /* Maximum possible length */
 		break;
@@ -158,7 +163,8 @@ static void at_sim_read_info(struct ofono_sim *sim, int fileid,
 				at_crsm_info_cb, cbd, g_free) > 0)
 		return;
 
-error:
+	g_free(cbd);
+
 	CALLBACK_WITH_FAILURE(cb, -1, -1, -1, NULL,
 				EF_STATUS_INVALIDATED, data);
 }
@@ -219,9 +225,6 @@ static void at_sim_read_binary(struct ofono_sim *sim, int fileid,
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 
-	if (cbd == NULL)
-		goto error;
-
 	snprintf(buf, sizeof(buf), "AT+CRSM=176,%i,%i,%i,%i", fileid,
 			start >> 8, start & 0xff, length);
 
@@ -229,7 +232,6 @@ static void at_sim_read_binary(struct ofono_sim *sim, int fileid,
 				at_crsm_read_cb, cbd, g_free) > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, NULL, 0, data);
@@ -243,9 +245,6 @@ static void at_sim_read_record(struct ofono_sim *sim, int fileid,
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[64];
 
-	if (cbd == NULL)
-		goto error;
-
 	snprintf(buf, sizeof(buf), "AT+CRSM=178,%i,%i,4,%i", fileid,
 			record, length);
 
@@ -253,7 +252,6 @@ static void at_sim_read_record(struct ofono_sim *sim, int fileid,
 				at_crsm_read_cb, cbd, g_free) > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, NULL, 0, data);
@@ -308,7 +306,7 @@ static void at_sim_update_binary(struct ofono_sim *sim, int fileid,
 	char *buf = g_try_new(char, 36 + length * 2);
 	int len, ret;
 
-	if (cbd == NULL || buf == NULL)
+	if (buf == NULL)
 		goto error;
 
 	len = sprintf(buf, "AT+CRSM=214,%i,%i,%i,%i,", fileid,
@@ -338,18 +336,29 @@ static void at_sim_update_record(struct ofono_sim *sim, int fileid,
 {
 	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
-	char *buf = g_try_new(char, 36 + length * 2);
+	char *buf;
 	int len, ret;
+	int size = 36 + length * 2;
 
-	if (cbd == NULL || buf == NULL)
+	if (sd->vendor == OFONO_VENDOR_MBM)
+		size += 2; /*Add quotes*/
+
+	buf = g_try_new(char, size);
+	if (buf == NULL)
 		goto error;
 
 	len = sprintf(buf, "AT+CRSM=220,%i,%i,4,%i,", fileid,
 			record, length);
 
+	if (sd->vendor == OFONO_VENDOR_MBM)
+		len += sprintf(buf + len, "\"");
+
 	for (; length; length--)
 		len += sprintf(buf + len, "%02hhX", *value++);
 
+	if (sd->vendor == OFONO_VENDOR_MBM)
+		sprintf(buf + len, "\"");
+
 	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_update_cb, cbd, g_free);
 
@@ -370,17 +379,28 @@ static void at_sim_update_cyclic(struct ofono_sim *sim, int fileid,
 {
 	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
-	char *buf = g_try_new(char, 36 + length * 2);
+	char *buf;
 	int len, ret;
+	int size = 36 + length * 2;
 
-	if (cbd == NULL || buf == NULL)
+	if (sd->vendor == OFONO_VENDOR_MBM)
+		size += 2; /* Add quotes */
+
+	buf = g_try_new(char, size);
+	if (buf == NULL)
 		goto error;
 
 	len = sprintf(buf, "AT+CRSM=220,%i,0,3,%i,", fileid, length);
 
+	if (sd->vendor == OFONO_VENDOR_MBM)
+		len += sprintf(buf + len, "\"");
+
 	for (; length; length--)
 		len += sprintf(buf + len, "%02hhX", *value++);
 
+	if (sd->vendor == OFONO_VENDOR_MBM)
+		sprintf(buf + len, "\"");
+
 	ret = g_at_chat_send(sd->chat, buf, crsm_prefix,
 				at_crsm_update_cb, cbd, g_free);
 
@@ -429,14 +449,10 @@ static void at_read_imsi(struct ofono_sim *sim, ofono_sim_imsi_cb_t cb,
 	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 
-	if (cbd == NULL)
-		goto error;
-
 	if (g_at_chat_send(sd->chat, "AT+CIMI", NULL,
 				at_cimi_cb, cbd, g_free) > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, NULL, data);
@@ -527,6 +543,44 @@ error:
 	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
 }
 
+static void zpinpuk_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	ofono_sim_pin_retries_cb_t cb = cbd->cb;
+	const char *final = g_at_result_final_response(result);
+	GAtResultIter iter;
+	struct ofono_error error;
+	int retries[OFONO_SIM_PASSWORD_INVALID];
+	size_t i;
+	static enum ofono_sim_password_type password_types[] = {
+		OFONO_SIM_PASSWORD_SIM_PIN,
+		OFONO_SIM_PASSWORD_SIM_PUK,
+	};
+
+
+	decode_at_error(&error, final);
+
+	if (!ok) {
+		cb(&error, NULL, cbd->data);
+		return;
+	}
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+ZPINPUK:"))
+		goto error;
+
+	BUILD_PIN_RETRIES_ARRAY(password_types, ARRAY_SIZE(password_types),
+				retries);
+
+	cb(&error, retries, cbd->data);
+
+	return;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
+}
+
 static void xpincnt_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
@@ -566,6 +620,148 @@ error:
 	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
 }
 
+static void cpnnum_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	ofono_sim_pin_retries_cb_t cb = cbd->cb;
+	const char *final = g_at_result_final_response(result);
+	GAtResultIter iter;
+	struct ofono_error error;
+	const char *line;
+	int num;
+	char **entries;
+	int retries[OFONO_SIM_PASSWORD_INVALID];
+	size_t i;
+
+	decode_at_error(&error, final);
+
+	if (!ok) {
+		cb(&error, NULL, cbd->data);
+		return;
+	}
+
+	g_at_result_iter_init(&iter, result);
+
+	for (num = 0; num < g_at_result_num_response_lines(result); num++)
+		g_at_result_iter_next(&iter, NULL);
+
+	line = g_at_result_iter_raw_line(&iter);
+
+	DBG("%s", line);
+
+	for (i = 0; i < OFONO_SIM_PASSWORD_INVALID; i++)
+		retries[i] = -1;
+
+	entries = g_strsplit(line, "; ", -1);
+
+	for (num = 0; entries[num]; num++) {
+		int retry;
+
+		if (strlen(entries[num]) < 5)
+			continue;
+
+		retry = strtol(entries[num] + 5, NULL, 10);
+		if (retry == 0 && errno == EINVAL)
+			continue;
+
+		if (g_str_has_prefix(entries[num], "PIN1=") == TRUE)
+			retries[OFONO_SIM_PASSWORD_SIM_PIN] = retry;
+		else if (g_str_has_prefix(entries[num], "PUK1=") == TRUE)
+			retries[OFONO_SIM_PASSWORD_SIM_PUK] = retry;
+		else if (g_str_has_prefix(entries[num], "PIN2=") == TRUE)
+			retries[OFONO_SIM_PASSWORD_SIM_PIN2] = retry;
+		else if (g_str_has_prefix(entries[num], "PUK2=") == TRUE)
+			retries[OFONO_SIM_PASSWORD_SIM_PUK2] = retry;
+	}
+
+	g_strfreev(entries);
+
+	cb(&error, retries, cbd->data);
+}
+
+static void at_epin_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	ofono_sim_pin_retries_cb_t cb = cbd->cb;
+	const char *final = g_at_result_final_response(result);
+	GAtResultIter iter;
+	struct ofono_error error;
+	int retries[OFONO_SIM_PASSWORD_INVALID];
+	size_t i;
+	static enum ofono_sim_password_type password_types[] = {
+		OFONO_SIM_PASSWORD_SIM_PIN,
+		OFONO_SIM_PASSWORD_SIM_PUK,
+		OFONO_SIM_PASSWORD_SIM_PIN2,
+		OFONO_SIM_PASSWORD_SIM_PUK2,
+	};
+
+	decode_at_error(&error, final);
+
+	if (!ok) {
+		cb(&error, NULL, cbd->data);
+		return;
+	}
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "*EPIN:"))
+		goto error;
+
+	BUILD_PIN_RETRIES_ARRAY(password_types, ARRAY_SIZE(password_types),
+				retries);
+
+	cb(&error, retries, cbd->data);
+
+	return;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
+}
+
+static void at_cpinr_cb(gboolean ok, GAtResult *result, gpointer user_data)
+{
+	struct cb_data *cbd = user_data;
+	ofono_sim_pin_retries_cb_t cb = cbd->cb;
+	GAtResultIter iter;
+	struct ofono_error error;
+	int retries[OFONO_SIM_PASSWORD_INVALID];
+	size_t len = sizeof(at_sim_name) / sizeof(*at_sim_name);
+	size_t i;
+
+	decode_at_error(&error, g_at_result_final_response(result));
+
+	if (!ok) {
+		cb(&error, NULL, cbd->data);
+		return;
+	}
+
+	for (i = 0; i < OFONO_SIM_PASSWORD_INVALID; i++)
+		retries[i] = -1;
+
+	g_at_result_iter_init(&iter, result);
+
+	/* Ignore +CPINRE results... */
+	while (g_at_result_iter_next(&iter, "+CPINR:")) {
+		const char *name;
+		int val;
+
+		if (!g_at_result_iter_next_unquoted_string(&iter, &name))
+			continue;
+
+		if (!g_at_result_iter_next_number(&iter, &val))
+			continue;
+
+		for (i = 1; i < len; i++) {
+			if (!strcmp(name, at_sim_name[i].name)) {
+				retries[i] = val;
+				break;
+			}
+		}
+	}
+
+	cb(&error, retries, cbd->data);
+}
+
 static void at_pin_retries_query(struct ofono_sim *sim,
 					ofono_sim_pin_retries_cb_t cb,
 					void *data)
@@ -575,27 +771,39 @@ static void at_pin_retries_query(struct ofono_sim *sim,
 
 	DBG("");
 
-	if (cbd == NULL)
-		goto error;
-
 	switch (sd->vendor) {
 	case OFONO_VENDOR_IFX:
 		if (g_at_chat_send(sd->chat, "AT+XPINCNT", xpincnt_prefix,
 					xpincnt_cb, cbd, g_free) > 0)
 			return;
-
+		break;
+	case OFONO_VENDOR_SPEEDUP:
+		if (g_at_chat_send(sd->chat, "AT+CPNNUM", NULL,
+					cpnnum_cb, cbd, g_free) > 0)
+			return;
 		break;
 	case OFONO_VENDOR_HUAWEI:
 		if (g_at_chat_send(sd->chat, "AT^CPIN?", huawei_cpin_prefix,
 					huawei_cpin_cb, cbd, g_free) > 0)
 			return;
-
+		break;
+	case OFONO_VENDOR_ZTE:
+		if (g_at_chat_send(sd->chat, "AT+ZPINPUK=?", zpinpuk_prefix,
+					zpinpuk_cb, cbd, g_free) > 0)
+			return;
+		break;
+	case OFONO_VENDOR_MBM:
+		if (g_at_chat_send(sd->chat, "AT*EPIN?", epin_prefix,
+					at_epin_cb, cbd, g_free) > 0)
+			return;
 		break;
 	default:
+		if (g_at_chat_send(sd->chat, "AT+CPINR", cpinr_prefixes,
+					at_cpinr_cb, cbd, g_free) > 0)
+			return;
 		break;
 	}
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, NULL, data);
@@ -662,16 +870,12 @@ static void at_pin_query(struct ofono_sim *sim, ofono_sim_passwd_cb_t cb,
 	struct sim_data *sd = ofono_sim_get_data(sim);
 	struct cb_data *cbd = cb_data_new(cb, data);
 
-	if (cbd == NULL)
-		goto error;
-
 	cbd->user = sim;
 
 	if (g_at_chat_send(sd->chat, "AT+CPIN?", cpin_prefix,
 				at_cpin_cb, cbd, g_free) > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, -1, data);
@@ -771,9 +975,6 @@ static void at_pin_send(struct ofono_sim *sim, const char *passwd,
 	char buf[64];
 	int ret;
 
-	if (cbd == NULL)
-		goto error;
-
 	cbd->user = sd;
 
 	snprintf(buf, sizeof(buf), "AT+CPIN=\"%s\"", passwd);
@@ -786,7 +987,6 @@ static void at_pin_send(struct ofono_sim *sim, const char *passwd,
 	if (ret > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
@@ -801,9 +1001,6 @@ static void at_pin_send_puk(struct ofono_sim *sim, const char *puk,
 	char buf[64];
 	int ret;
 
-	if (cbd == NULL)
-		goto error;
-
 	cbd->user = sd;
 
 	snprintf(buf, sizeof(buf), "AT+CPIN=\"%s\",\"%s\"", puk, passwd);
@@ -816,7 +1013,6 @@ static void at_pin_send_puk(struct ofono_sim *sim, const char *puk,
 	if (ret > 0)
 		return;
 
-error:
 	g_free(cbd);
 
 	CALLBACK_WITH_FAILURE(cb, data);
@@ -856,9 +1052,6 @@ static void at_pin_enable(struct ofono_sim *sim,
 	int ret;
 	unsigned int len = sizeof(at_clck_cpwd_fac) / sizeof(*at_clck_cpwd_fac);
 
-	if (cbd == NULL)
-		goto error;
-
 	if (passwd_type >= len || at_clck_cpwd_fac[passwd_type] == NULL)
 		goto error;
 
@@ -881,7 +1074,7 @@ error:
 
 static void at_change_passwd(struct ofono_sim *sim,
 			enum ofono_sim_password_type passwd_type,
-			const char *old, const char *new,
+			const char *old_passwd, const char *new_passwd,
 			ofono_sim_lock_unlock_cb_t cb, void *data)
 {
 	struct sim_data *sd = ofono_sim_get_data(sim);
@@ -890,15 +1083,12 @@ static void at_change_passwd(struct ofono_sim *sim,
 	int ret;
 	unsigned int len = sizeof(at_clck_cpwd_fac) / sizeof(*at_clck_cpwd_fac);
 
-	if (cbd == NULL)
-		goto error;
-
 	if (passwd_type >= len ||
 			at_clck_cpwd_fac[passwd_type] == NULL)
 		goto error;
 
 	snprintf(buf, sizeof(buf), "AT+CPWD=\"%s\",\"%s\",\"%s\"",
-			at_clck_cpwd_fac[passwd_type], old, new);
+			at_clck_cpwd_fac[passwd_type], old_passwd, new_passwd);
 
 	ret = g_at_chat_send(sd->chat, buf, none_prefix,
 				at_lock_unlock_cb, cbd, g_free);
@@ -953,9 +1143,6 @@ static void at_pin_query_enabled(struct ofono_sim *sim,
 	char buf[64];
 	unsigned int len = sizeof(at_clck_cpwd_fac) / sizeof(*at_clck_cpwd_fac);
 
-	if (cbd == NULL)
-		goto error;
-
 	if (passwd_type >= len || at_clck_cpwd_fac[passwd_type] == NULL)
 		goto error;
 
