@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -54,15 +54,17 @@
 #include <ofono/sim.h>
 #include <ofono/stk.h>
 #include <ofono/sms.h>
-#include <ofono/ssn.h>
 #include <ofono/ussd.h>
 #include <ofono/voicecall.h>
 #include <ofono/gprs.h>
 #include <ofono/gprs-context.h>
+#include <ofono/gnss.h>
 
 #include <drivers/atmodem/vendor.h>
-#include <drivers/atmodem/sim-poll.h>
 #include <drivers/atmodem/atutil.h>
+#include <drivers/hfpmodem/slc.h>
+
+#include "ofono.h"
 
 static const char *none_prefix[] = { NULL };
 static const char *ptty_prefix[] = { "+PTTY:", NULL };
@@ -73,24 +75,45 @@ struct phonesim_data {
 	GAtChat *chat;
 	gboolean calypso;
 	gboolean use_mux;
+	gboolean hfp;
+	struct hfp_slc_info hfp_info;
+	unsigned int hfp_watch;
+	int batt_level;
 };
 
 struct gprs_context_data {
 	GAtChat *chat;
 	char *interface;
+	enum ofono_gprs_proto proto;
 };
 
 static void at_cgact_up_cb(gboolean ok, GAtResult *result, gpointer user_data)
 {
 	struct cb_data *cbd = user_data;
-	ofono_gprs_context_up_cb_t cb = cbd->cb;
+	ofono_gprs_context_cb_t cb = cbd->cb;
 	struct ofono_gprs_context *gc = cbd->user;
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 	struct ofono_error error;
 
 	decode_at_error(&error, g_at_result_final_response(result));
-	cb(&error, ok ? gcd->interface : NULL, FALSE,
-			NULL, NULL, NULL, NULL, cbd->data);
+
+	if (ok == FALSE)
+		goto done;
+
+	ofono_gprs_context_set_interface(gc, gcd->interface);
+
+	if (gcd->proto == OFONO_GPRS_PROTO_IP ||
+			gcd->proto == OFONO_GPRS_PROTO_IPV4V6)
+		ofono_gprs_context_set_ipv4_address(gc, NULL, FALSE);
+
+	if (gcd->proto == OFONO_GPRS_PROTO_IPV6 ||
+			gcd->proto == OFONO_GPRS_PROTO_IPV4V6) {
+		ofono_gprs_context_set_ipv6_address(gc, "fe80::1");
+		ofono_gprs_context_set_ipv6_prefix_length(gc, 10);
+	}
+
+done:
+	cb(&error, cbd->data);
 }
 
 static void at_cgact_down_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -105,16 +128,32 @@ static void at_cgact_down_cb(gboolean ok, GAtResult *result, gpointer user_data)
 
 static void phonesim_activate_primary(struct ofono_gprs_context *gc,
 				const struct ofono_gprs_primary_context *ctx,
-				ofono_gprs_context_up_cb_t cb, void *data)
+				ofono_gprs_context_cb_t cb, void *data)
 {
 	struct gprs_context_data *gcd = ofono_gprs_context_get_data(gc);
 	struct cb_data *cbd = cb_data_new(cb, data);
 	char buf[OFONO_GPRS_MAX_APN_LENGTH + 128];
-	int len;
+	int len = 0;
 
 	cbd->user = gc;
+	gcd->proto = ctx->proto;
 
-	len = snprintf(buf, sizeof(buf), "AT+CGDCONT=%u,\"IP\"", ctx->cid);
+	switch (ctx->proto) {
+	case OFONO_GPRS_PROTO_IP:
+		len = snprintf(buf, sizeof(buf), "AT+CGDCONT=%u,\"IP\"",
+				ctx->cid);
+		break;
+
+	case OFONO_GPRS_PROTO_IPV6:
+		len = snprintf(buf, sizeof(buf), "AT+CGDCONT=%u,\"IPV6\"",
+				ctx->cid);
+		break;
+
+	case OFONO_GPRS_PROTO_IPV4V6:
+		len = snprintf(buf, sizeof(buf), "AT+CGDCONT=%u,\"IPV4V6\"",
+				ctx->cid);
+		break;
+	}
 
 	if (ctx->apn)
 		snprintf(buf + len, sizeof(buf) - len - 3, ",\"%s\"",
@@ -132,7 +171,7 @@ static void phonesim_activate_primary(struct ofono_gprs_context *gc,
 error:
 	g_free(cbd);
 
-	CALLBACK_WITH_FAILURE(cb, NULL, 0, NULL, NULL, NULL, NULL, data);
+	CALLBACK_WITH_FAILURE(cb, data);
 }
 
 static void phonesim_deactivate_primary(struct ofono_gprs_context *gc,
@@ -308,11 +347,11 @@ static void phonesim_ctm_set(struct ofono_ctm *ctm, ofono_bool_t enable,
 }
 
 static struct ofono_gprs_context_driver context_driver = {
-	.name                   = "phonesim",
-	.probe                  = phonesim_context_probe,
-	.remove                 = phonesim_context_remove,
-	.activate_primary       = phonesim_activate_primary,
-	.deactivate_primary     = phonesim_deactivate_primary,
+	.name			= "phonesim",
+	.probe			= phonesim_context_probe,
+	.remove			= phonesim_context_remove,
+	.activate_primary	= phonesim_activate_primary,
+	.deactivate_primary	= phonesim_deactivate_primary,
 };
 
 static struct ofono_ctm_driver ctm_driver = {
@@ -348,9 +387,9 @@ static void phonesim_remove(struct ofono_modem *modem)
 	ofono_modem_set_data(modem, NULL);
 }
 
-static void phonesim_debug(const char *str, void *user_data)
+static void phonesim_debug(const char *str, void *prefix)
 {
-	ofono_info("%s", str);
+	ofono_info("%s%s", (const char *) prefix, str);
 }
 
 static void cfun_set_on_cb(gboolean ok, GAtResult *result, gpointer user_data)
@@ -386,6 +425,43 @@ static void crst_notify(GAtResult *result, gpointer user_data)
 	g_idle_add(phonesim_reset, user_data);
 }
 
+static void emulator_battery_cb(struct ofono_atom *atom, void *data)
+{
+	struct ofono_emulator *em = __ofono_atom_get_data(atom);
+	int val = 0;
+
+	if (GPOINTER_TO_INT(data) > 0)
+		val = (GPOINTER_TO_INT(data) - 1) / 20 + 1;
+
+	ofono_emulator_set_indicator(em, OFONO_EMULATOR_IND_BATTERY, val);
+}
+
+static void cbc_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_modem *modem = user_data;
+	struct phonesim_data *data = ofono_modem_get_data(modem);
+	GAtResultIter iter;
+	int status;
+	int level;
+
+	g_at_result_iter_init(&iter, result);
+	if (!g_at_result_iter_next(&iter, "+CBC:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &status))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &level))
+		return;
+
+	data->batt_level = level;
+
+	__ofono_modem_foreach_registered_atom(modem,
+						OFONO_ATOM_TYPE_EMULATOR_HFP,
+						emulator_battery_cb,
+						GUINT_TO_POINTER(level));
+}
+
 static void phonesim_disconnected(gpointer user_data)
 {
 	struct ofono_modem *modem = user_data;
@@ -422,7 +498,7 @@ static void mux_setup(GAtMux *mux, gpointer user_data)
 	data->mux = mux;
 
 	if (getenv("OFONO_AT_DEBUG"))
-		g_at_mux_set_debug(data->mux, phonesim_debug, NULL);
+		g_at_mux_set_debug(data->mux, phonesim_debug, "");
 
 	g_at_mux_start(mux);
 	io = g_at_mux_create_channel(mux);
@@ -437,7 +513,7 @@ static void mux_setup(GAtMux *mux, gpointer user_data)
 	g_io_channel_unref(io);
 
 	if (getenv("OFONO_AT_DEBUG"))
-		g_at_chat_set_debug(data->chat, phonesim_debug, NULL);
+		g_at_chat_set_debug(data->chat, phonesim_debug, "");
 
 	if (data->calypso)
 		g_at_chat_set_wakeup_command(data->chat, "AT\r", 500, 5000);
@@ -448,14 +524,49 @@ static void mux_setup(GAtMux *mux, gpointer user_data)
 					cfun_set_on_cb, modem, NULL);
 }
 
+static void emulator_hfp_watch(struct ofono_atom *atom,
+				enum ofono_atom_watch_condition cond,
+				void *user_data)
+{
+	struct phonesim_data *data = user_data;
+
+	if (cond != OFONO_ATOM_WATCH_CONDITION_REGISTERED)
+		return;
+
+	emulator_battery_cb(atom, GUINT_TO_POINTER(data->batt_level));
+}
+
+static int connect_socket(const char *address, int port)
+{
+	struct sockaddr_in addr;
+	int sk;
+	int err;
+
+	sk = socket(PF_INET, SOCK_STREAM, 0);
+	if (sk < 0)
+		return -EINVAL;
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sin_family = AF_INET;
+	addr.sin_addr.s_addr = inet_addr(address);
+	addr.sin_port = htons(port);
+
+	err = connect(sk, (struct sockaddr *) &addr, sizeof(addr));
+	if (err < 0) {
+		close(sk);
+		return -errno;
+	}
+
+	return sk;
+}
+
 static int phonesim_enable(struct ofono_modem *modem)
 {
 	struct phonesim_data *data = ofono_modem_get_data(modem);
 	GIOChannel *io;
 	GAtSyntax *syntax;
-	struct sockaddr_in addr;
 	const char *address, *value;
-	int sk, err, port;
+	int sk, port;
 
 	DBG("%p", modem);
 
@@ -475,20 +586,9 @@ static int phonesim_enable(struct ofono_modem *modem)
 	if (!g_strcmp0(value, "internal"))
 		data->use_mux = TRUE;
 
-	sk = socket(PF_INET, SOCK_STREAM, 0);
+	sk = connect_socket(address, port);
 	if (sk < 0)
-		return -EINVAL;
-
-	memset(&addr, 0, sizeof(addr));
-	addr.sin_family = AF_INET;
-	addr.sin_addr.s_addr = inet_addr(address);
-	addr.sin_port = htons(port);
-
-	err = connect(sk, (struct sockaddr *) &addr, sizeof(addr));
-	if (err < 0) {
-		close(sk);
-		return err;
-	}
+		return sk;
 
 	io = g_io_channel_unix_new(sk);
 	if (io == NULL) {
@@ -510,7 +610,7 @@ static int phonesim_enable(struct ofono_modem *modem)
 		return -ENOMEM;
 
 	if (getenv("OFONO_AT_DEBUG"))
-		g_at_chat_set_debug(data->chat, phonesim_debug, NULL);
+		g_at_chat_set_debug(data->chat, phonesim_debug, "");
 
 	g_at_chat_set_disconnect_function(data->chat,
 						phonesim_disconnected, modem);
@@ -541,6 +641,15 @@ static int phonesim_enable(struct ofono_modem *modem)
 	g_at_chat_register(data->chat, "+CRST:",
 				crst_notify, FALSE, modem, NULL);
 
+	g_at_chat_register(data->chat, "+CBC:",
+				cbc_notify, FALSE, modem, NULL);
+
+	g_at_chat_send(data->chat, "AT+CBC", none_prefix, NULL, NULL, NULL);
+
+	data->hfp_watch = __ofono_modem_add_atom_watch(modem,
+					OFONO_ATOM_TYPE_EMULATOR_HFP,
+					emulator_hfp_watch, data, NULL);
+
 	return 0;
 }
 
@@ -579,6 +688,8 @@ static int phonesim_disable(struct ofono_modem *modem)
 
 	DBG("%p", modem);
 
+	__ofono_modem_remove_atom_watch(modem, data->hfp_watch);
+
 	g_at_chat_unref(data->chat);
 	data->chat = NULL;
 
@@ -651,19 +762,18 @@ static void phonesim_post_online(struct ofono_modem *modem)
 
 	ofono_call_meter_create(modem, 0, "atmodem", data->chat);
 	ofono_call_barring_create(modem, 0, "atmodem", data->chat);
-	ofono_ssn_create(modem, 0, "atmodem", data->chat);
 	ofono_call_volume_create(modem, 0, "atmodem", data->chat);
 
 	if (!data->calypso)
 		ofono_cbs_create(modem, 0, "atmodem", data->chat);
 
-	gprs = ofono_gprs_create(modem, 0, "atmodem", data->chat);
-
 	gc1 = ofono_gprs_context_create(modem, 0, "phonesim", data->chat);
+	gprs = ofono_gprs_create(modem, 0, "atmodem", data->chat);
+	gc2 = ofono_gprs_context_create(modem, 0, "phonesim", data->chat);
+
 	if (gprs && gc1)
 		ofono_gprs_add_context(gprs, gc1);
 
-	gc2 = ofono_gprs_context_create(modem, 0, "phonesim", data->chat);
 	if (gprs && gc2)
 		ofono_gprs_add_context(gprs, gc2);
 
@@ -671,6 +781,7 @@ static void phonesim_post_online(struct ofono_modem *modem)
 	if (mw)
 		ofono_message_waiting_register(mw);
 
+	ofono_gnss_create(modem, 0, "atmodem", data->chat);
 }
 
 static struct ofono_modem_driver phonesim_driver = {
@@ -685,14 +796,142 @@ static struct ofono_modem_driver phonesim_driver = {
 	.post_online	= phonesim_post_online,
 };
 
+static int localhfp_probe(struct ofono_modem *modem)
+{
+	struct hfp_slc_info *info;
+
+	DBG("%p", modem);
+
+	info = g_try_new(struct hfp_slc_info, 1);
+	if (info == NULL)
+		return -ENOMEM;
+
+	ofono_modem_set_data(modem, info);
+
+	return 0;
+}
+
+static void localhfp_remove(struct ofono_modem *modem)
+{
+	struct hfp_slc_info *info = ofono_modem_get_data(modem);
+
+	DBG("%p", modem);
+
+	g_free(info);
+	ofono_modem_set_data(modem, NULL);
+}
+
+static void slc_established(gpointer userdata)
+{
+	struct ofono_modem *modem = userdata;
+
+	ofono_modem_set_powered(modem, TRUE);
+}
+
+static void slc_failed(gpointer userdata)
+{
+	struct ofono_modem *modem = userdata;
+	struct hfp_slc_info *info = ofono_modem_get_data(modem);
+
+	ofono_modem_set_powered(modem, FALSE);
+
+	g_at_chat_unref(info->chat);
+	info->chat = NULL;
+}
+
+static int localhfp_enable(struct ofono_modem *modem)
+{
+	struct hfp_slc_info *info = ofono_modem_get_data(modem);
+	GIOChannel *io;
+	GAtSyntax *syntax;
+	GAtChat *chat;
+	const char *address;
+	int sk, port;
+
+	address = ofono_modem_get_string(modem, "Address");
+	if (address == NULL)
+		return -EINVAL;
+
+	port = ofono_modem_get_integer(modem, "Port");
+	if (port < 0)
+		return -EINVAL;
+
+	sk = connect_socket(address, port);
+	if (sk < 0)
+		return sk;
+
+	io = g_io_channel_unix_new(sk);
+	if (io == NULL) {
+		close(sk);
+		return -ENOMEM;
+	}
+
+	syntax = g_at_syntax_new_gsmv1();
+	chat = g_at_chat_new(io, syntax);
+	g_at_syntax_unref(syntax);
+	g_io_channel_unref(io);
+
+	if (chat == NULL)
+		return -ENOMEM;
+
+	if (getenv("OFONO_AT_DEBUG"))
+		g_at_chat_set_debug(chat, phonesim_debug, "LocalHfp: ");
+
+	g_at_chat_set_disconnect_function(chat, slc_failed, modem);
+
+	hfp_slc_info_init(info, HFP_VERSION_LATEST);
+	info->chat = chat;
+	hfp_slc_establish(info, slc_established, slc_failed, modem);
+
+	return -EINPROGRESS;
+}
+
+static int localhfp_disable(struct ofono_modem *modem)
+{
+	struct hfp_slc_info *info = ofono_modem_get_data(modem);
+
+	g_at_chat_unref(info->chat);
+	info->chat = NULL;
+
+	return 0;
+}
+
+static void localhfp_pre_sim(struct ofono_modem *modem)
+{
+	struct hfp_slc_info *info = ofono_modem_get_data(modem);
+
+	DBG("%p", modem);
+
+	ofono_voicecall_create(modem, 0, "hfpmodem", info);
+	ofono_netreg_create(modem, 0, "hfpmodem", info);
+	ofono_call_volume_create(modem, 0, "hfpmodem", info);
+}
+
+static struct ofono_modem_driver localhfp_driver = {
+	.name		= "localhfp",
+	.probe		= localhfp_probe,
+	.remove		= localhfp_remove,
+	.enable		= localhfp_enable,
+	.disable	= localhfp_disable,
+	.pre_sim	= localhfp_pre_sim,
+};
+
 static struct ofono_modem *create_modem(GKeyFile *keyfile, const char *group)
 {
+	const char *driver = "phonesim";
 	struct ofono_modem *modem;
 	char *value;
 
 	DBG("group %s", group);
 
-	modem = ofono_modem_create(group, "phonesim");
+	value = g_key_file_get_string(keyfile, group, "Modem", NULL);
+
+	if (value && g_str_equal(value, "hfp"))
+		driver = "localhfp";
+
+	g_free(value);
+
+	modem = ofono_modem_create(group, driver);
 	if (modem == NULL)
 		return NULL;
 
@@ -783,8 +1022,9 @@ static int phonesim_init(void)
 	if (err < 0)
 		return err;
 
-	ofono_gprs_context_driver_register(&context_driver);
+	ofono_modem_driver_register(&localhfp_driver);
 
+	ofono_gprs_context_driver_register(&context_driver);
 	ofono_ctm_driver_register(&ctm_driver);
 
 	parse_config(CONFIGDIR "/phonesim.conf");
