@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -42,21 +42,16 @@
 #define MAX_VOICE_CALLS 16
 
 #define VOICECALL_FLAG_SIM_ECC_READY 0x1
+#define VOICECALL_FLAG_STK_MODEM_CALLSETUP 0x2
 
 #define SETTINGS_STORE "voicecall"
 #define SETTINGS_GROUP "Settings"
 
 GSList *g_drivers = NULL;
 
-struct multi_release {
-	ofono_voicecall_cb_t cb;
-	void *data;
-};
-
 struct ofono_voicecall {
 	GSList *call_list;
 	GSList *release_list;
-	struct multi_release multi_release;
 	GSList *multiparty_list;
 	GHashTable *en_list; /* emergency number list */
 	GSList *sim_en_list; /* Emergency numbers already read from SIM */
@@ -77,6 +72,9 @@ struct ofono_voicecall {
 	unsigned int hfp_watch;
 	GKeyFile *settings;
 	char *imsi;
+	ofono_voicecall_cb_t release_queue_done_cb;
+	struct ofono_emulator *pending_em;
+	unsigned int pending_id;
 };
 
 struct voicecall {
@@ -112,6 +110,11 @@ struct tone_queue_entry {
 	int id;
 };
 
+struct emulator_status {
+	struct ofono_voicecall *vc;
+	int status;
+};
+
 static const char *default_en_list[] = { "911", "112", NULL };
 static const char *default_en_list_no_sim[] = { "119", "118", "999", "110",
 						"08", "000", NULL };
@@ -259,6 +262,54 @@ static unsigned int voicecalls_num_connecting(struct ofono_voicecall *vc)
 	return r;
 }
 
+static gboolean voicecalls_have_active(struct ofono_voicecall *vc)
+{
+	GSList *l;
+	struct voicecall *v;
+
+	for (l = vc->call_list; l; l = l->next) {
+		v = l->data;
+
+		if (v->call->status == CALL_STATUS_ACTIVE ||
+				v->call->status == CALL_STATUS_DIALING ||
+				v->call->status == CALL_STATUS_ALERTING)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean voicecalls_have_with_status(struct ofono_voicecall *vc,
+						int status)
+{
+	GSList *l;
+	struct voicecall *v;
+
+	for (l = vc->call_list; l; l = l->next) {
+		v = l->data;
+
+		if (v->call->status == status)
+			return TRUE;
+	}
+
+	return FALSE;
+}
+
+static gboolean voicecalls_have_held(struct ofono_voicecall *vc)
+{
+	return voicecalls_have_with_status(vc, CALL_STATUS_HELD);
+}
+
+static gboolean voicecalls_have_waiting(struct ofono_voicecall *vc)
+{
+	return voicecalls_have_with_status(vc, CALL_STATUS_WAITING);
+}
+
+static gboolean voicecalls_have_incoming(struct ofono_voicecall *vc)
+{
+	return voicecalls_have_with_status(vc, CALL_STATUS_INCOMING);
+}
+
 static void dial_request_finish(struct ofono_voicecall *vc)
 {
 	struct dial_request *dial_req = vc->dial_req;
@@ -469,13 +520,13 @@ static DBusMessage *voicecall_deflect(DBusConnection *conn,
 	const char *number;
 
 	if (call->status != CALL_STATUS_INCOMING &&
-		call->status != CALL_STATUS_WAITING)
+			call->status != CALL_STATUS_WAITING)
 		return __ofono_error_failed(msg);
 
 	if (vc->driver->deflect == NULL)
 		return __ofono_error_not_implemented(msg);
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &number,
@@ -494,16 +545,6 @@ static DBusMessage *voicecall_deflect(DBusConnection *conn,
 	return NULL;
 }
 
-static void dial_request_user_cancel(struct ofono_voicecall *vc,
-					struct voicecall *call)
-{
-	if (vc->dial_req == NULL)
-		return;
-
-	if (call == NULL || call == vc->dial_req->call)
-		dial_request_finish(vc);
-}
-
 static DBusMessage *voicecall_hangup(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
@@ -512,10 +553,11 @@ static DBusMessage *voicecall_hangup(DBusConnection *conn,
 	struct ofono_call *call = v->call;
 	gboolean single_call = vc->call_list->next == 0;
 
-	if (vc->pending)
+	if (vc->pending || vc->pending_em)
 		return __ofono_error_busy(msg);
 
-	dial_request_user_cancel(vc, v);
+	if (vc->dial_req && vc->dial_req->call != v)
+		return __ofono_error_busy(msg);
 
 	switch (call->status) {
 	case CALL_STATUS_DISCONNECTED:
@@ -545,7 +587,9 @@ static DBusMessage *voicecall_hangup(DBusConnection *conn,
 		return NULL;
 
 	case CALL_STATUS_HELD:
-		if (single_call && vc->driver->release_all_held) {
+		if (vc->driver->release_all_held &&
+				voicecalls_num_held(vc) == 1 &&
+				voicecalls_have_waiting(vc) == FALSE) {
 			vc->pending = dbus_message_ref(msg);
 			vc->driver->release_all_held(vc, generic_callback, vc);
 
@@ -610,7 +654,7 @@ static DBusMessage *voicecall_answer(DBusConnection *conn,
 	if (vc->driver->answer == NULL)
 		return __ofono_error_not_implemented(msg);
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	vc->pending = dbus_message_ref(msg);
@@ -704,32 +748,35 @@ static void voicecall_emit_multiparty(struct voicecall *call, gboolean mpty)
 static void emulator_call_status_cb(struct ofono_atom *atom, void *data)
 {
 	struct ofono_emulator *em = __ofono_atom_get_data(atom);
+	struct emulator_status *s = data;
 
-	ofono_emulator_set_indicator(em, OFONO_EMULATOR_IND_CALL,
-						GPOINTER_TO_INT(data));
+	ofono_emulator_set_indicator(em, OFONO_EMULATOR_IND_CALL, s->status);
 }
 
 static void emulator_callsetup_status_cb(struct ofono_atom *atom, void *data)
 {
 	struct ofono_emulator *em = __ofono_atom_get_data(atom);
+	struct emulator_status *s = data;
 
 	ofono_emulator_set_indicator(em, OFONO_EMULATOR_IND_CALLSETUP,
-						GPOINTER_TO_INT(data));
+					s->status);
 }
 
 static void emulator_callheld_status_cb(struct ofono_atom *atom, void *data)
 {
 	struct ofono_emulator *em = __ofono_atom_get_data(atom);
+	struct emulator_status *s = data;
 
 	ofono_emulator_set_indicator(em, OFONO_EMULATOR_IND_CALLHELD,
-						GPOINTER_TO_INT(data));
+					s->status);
 }
 
 static void notify_emulator_call_status(struct ofono_voicecall *vc)
 {
 	struct ofono_modem *modem = __ofono_atom_get_modem(vc->atom);
-	int status;
 	gboolean call = FALSE;
+	unsigned int non_mpty = 0;
+	gboolean multiparty = FALSE;
 	gboolean held = FALSE;
 	gboolean incoming = FALSE;
 	gboolean dialing = FALSE;
@@ -737,6 +784,9 @@ static void notify_emulator_call_status(struct ofono_voicecall *vc)
 	gboolean waiting = FALSE;
 	GSList *l;
 	struct voicecall *v;
+	struct emulator_status data;
+
+	data.vc = vc;
 
 	for (l = vc->call_list; l; l = l->next) {
 		v = l->data;
@@ -744,6 +794,12 @@ static void notify_emulator_call_status(struct ofono_voicecall *vc)
 		switch (v->call->status) {
 		case CALL_STATUS_ACTIVE:
 			call = TRUE;
+			if (g_slist_find_custom(vc->multiparty_list,
+						GINT_TO_POINTER(v->call->id),
+						call_compare_by_id))
+				multiparty = TRUE;
+			else
+				non_mpty++;
 			break;
 
 		case CALL_STATUS_HELD:
@@ -768,40 +824,51 @@ static void notify_emulator_call_status(struct ofono_voicecall *vc)
 		}
 	}
 
-	status = call || held ? OFONO_EMULATOR_CALL_ACTIVE :
+	data.status = call || held ? OFONO_EMULATOR_CALL_ACTIVE :
 					OFONO_EMULATOR_CALL_INACTIVE;
 
 	__ofono_modem_foreach_registered_atom(modem,
 						OFONO_ATOM_TYPE_EMULATOR_HFP,
-						emulator_call_status_cb,
-						GINT_TO_POINTER(status));
+						emulator_call_status_cb, &data);
 
 	if (incoming)
-		status = OFONO_EMULATOR_CALLSETUP_INCOMING;
+		data.status = OFONO_EMULATOR_CALLSETUP_INCOMING;
 	else if (dialing)
-		status = OFONO_EMULATOR_CALLSETUP_OUTGOING;
+		data.status = OFONO_EMULATOR_CALLSETUP_OUTGOING;
 	else if (alerting)
-		status = OFONO_EMULATOR_CALLSETUP_ALERTING;
+		data.status = OFONO_EMULATOR_CALLSETUP_ALERTING;
 	else if (waiting)
-		status = OFONO_EMULATOR_CALLSETUP_INCOMING;
+		data.status = OFONO_EMULATOR_CALLSETUP_INCOMING;
 	else
-		status = OFONO_EMULATOR_CALLSETUP_INACTIVE;
+		data.status = OFONO_EMULATOR_CALLSETUP_INACTIVE;
 
 	__ofono_modem_foreach_registered_atom(modem,
 						OFONO_ATOM_TYPE_EMULATOR_HFP,
 						emulator_callsetup_status_cb,
-						GINT_TO_POINTER(status));
+						&data);
 
 	if (held)
-		status = call ? OFONO_EMULATOR_CALLHELD_MULTIPLE :
+		data.status = call ? OFONO_EMULATOR_CALLHELD_MULTIPLE :
 					OFONO_EMULATOR_CALLHELD_ON_HOLD;
+	else if (non_mpty > 1 || (non_mpty && multiparty))
+		/*
+		 * After call swap, it is possible that all calls move
+		 * temporarily to active state (depending on call state update
+		 * order), generating an update of callheld indicator to 0.
+		 * This will fail PTS test TP/TWC/BV-03-I.
+		 *
+		 * So, in case of multiple active calls, or an active call with
+		 * an active mutiparty call, force update of callheld indicator
+		 * to 2 (intermediate state allowed).
+		 */
+		data.status = OFONO_EMULATOR_CALLHELD_ON_HOLD;
 	else
-		status = OFONO_EMULATOR_CALLHELD_NONE;
+		data.status = OFONO_EMULATOR_CALLHELD_NONE;
 
 	__ofono_modem_foreach_registered_atom(modem,
 						OFONO_ATOM_TYPE_EMULATOR_HFP,
 						emulator_callheld_status_cb,
-						GINT_TO_POINTER(status));
+						&data);
 }
 
 static void voicecall_set_call_status(struct voicecall *call, int status)
@@ -1023,44 +1090,6 @@ static int voicecalls_path_list(struct ofono_voicecall *vc, GSList *call_list,
 	return 0;
 }
 
-static gboolean voicecalls_have_active(struct ofono_voicecall *vc)
-{
-	GSList *l;
-	struct voicecall *v;
-
-	for (l = vc->call_list; l; l = l->next) {
-		v = l->data;
-
-		if (v->call->status == CALL_STATUS_ACTIVE ||
-				v->call->status == CALL_STATUS_DIALING ||
-				v->call->status == CALL_STATUS_ALERTING)
-			return TRUE;
-	}
-
-	return FALSE;
-}
-
-static gboolean voicecalls_have_with_status(struct ofono_voicecall *vc,
-						int status)
-{
-	GSList *l;
-	struct voicecall *v;
-
-	for (l = vc->call_list; l; l = l->next) {
-		v = l->data;
-
-		if (v->call->status == status)
-			return TRUE;
-	}
-
-	return FALSE;
-}
-
-static gboolean voicecalls_have_held(struct ofono_voicecall *vc)
-{
-	return voicecalls_have_with_status(vc, CALL_STATUS_HELD);
-}
-
 static GSList *voicecalls_held_list(struct ofono_voicecall *vc)
 {
 	GSList *l;
@@ -1103,16 +1132,6 @@ static GSList *voicecalls_active_list(struct ofono_voicecall *vc)
 	return r;
 }
 
-static gboolean voicecalls_have_waiting(struct ofono_voicecall *vc)
-{
-	return voicecalls_have_with_status(vc, CALL_STATUS_WAITING);
-}
-
-static gboolean voicecalls_have_incoming(struct ofono_voicecall *vc)
-{
-	return voicecalls_have_with_status(vc, CALL_STATUS_INCOMING);
-}
-
 struct ofono_call *__ofono_voicecall_find_call_with_status(
 				struct ofono_voicecall *vc, int status)
 {
@@ -1199,7 +1218,8 @@ static void voicecalls_emit_call_added(struct ofono_voicecall *vc,
 	g_dbus_send_message(ofono_dbus_get_connection(), signal);
 }
 
-static void voicecalls_release_queue(struct ofono_voicecall *vc, GSList *calls)
+static void voicecalls_release_queue(struct ofono_voicecall *vc, GSList *calls,
+					ofono_voicecall_cb_t cb)
 {
 	GSList *l;
 	struct voicecall *call;
@@ -1215,6 +1235,8 @@ static void voicecalls_release_queue(struct ofono_voicecall *vc, GSList *calls)
 
 		vc->release_list = g_slist_prepend(vc->release_list, l->data);
 	}
+
+	vc->release_queue_done_cb = cb;
 }
 
 static void voicecalls_release_next(struct ofono_voicecall *vc)
@@ -1477,14 +1499,14 @@ static int voicecall_dial(struct ofono_voicecall *vc, const char *number,
 
 	string_to_phone_number(number, &ph);
 
-	vc->driver->dial(vc, &ph, clir, cb, vc);
-
 	if (vc->settings) {
 		g_key_file_set_string(vc->settings, SETTINGS_GROUP,
 					"Number", number);
 		storage_sync(vc->imsi, SETTINGS_STORE, vc->settings);
 	}
 
+	vc->driver->dial(vc, &ph, clir, cb, vc);
+
 	return 0;
 }
 
@@ -1497,7 +1519,7 @@ static DBusMessage *manager_dial(DBusConnection *conn,
 	enum ofono_clir_option clir;
 	int err;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &number,
@@ -1539,7 +1561,7 @@ static DBusMessage *manager_transfer(DBusConnection *conn,
 	int numactive;
 	int numheld;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	numactive = voicecalls_num_active(vc);
@@ -1571,7 +1593,7 @@ static DBusMessage *manager_swap_without_accept(DBusConnection *conn,
 {
 	struct ofono_voicecall *vc = data;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	vc->pending = dbus_message_ref(msg);
@@ -1590,7 +1612,7 @@ static DBusMessage *manager_swap_calls(DBusConnection *conn,
 	if (vc->driver->swap_without_accept)
 		return manager_swap_without_accept(conn, msg, data);
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (voicecalls_have_waiting(vc))
@@ -1611,7 +1633,7 @@ static DBusMessage *manager_release_and_answer(DBusConnection *conn,
 {
 	struct ofono_voicecall *vc = data;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (!voicecalls_have_waiting(vc))
@@ -1632,7 +1654,7 @@ static DBusMessage *manager_hold_and_answer(DBusConnection *conn,
 {
 	struct ofono_voicecall *vc = data;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (voicecalls_have_waiting(vc) == FALSE)
@@ -1660,7 +1682,10 @@ static DBusMessage *manager_hangup_all(DBusConnection *conn,
 {
 	struct ofono_voicecall *vc = data;
 
-	if (vc->pending || vc->release_list)
+	if (vc->pending || vc->pending_em)
+		return __ofono_error_busy(msg);
+
+	if (vc->dial_req && vc->dial_req->call == NULL)
 		return __ofono_error_busy(msg);
 
 	if (vc->driver->hangup_all == NULL &&
@@ -1676,15 +1701,12 @@ static DBusMessage *manager_hangup_all(DBusConnection *conn,
 	vc->pending = dbus_message_ref(msg);
 
 	if (vc->driver->hangup_all == NULL) {
-		voicecalls_release_queue(vc, vc->call_list);
-		vc->multi_release.cb = voicecalls_release_done;
-		vc->multi_release.data = vc;
+		voicecalls_release_queue(vc, vc->call_list,
+						voicecalls_release_done);
 		voicecalls_release_next(vc);
 	} else
 		vc->driver->hangup_all(vc, generic_callback, vc);
 
-	dial_request_user_cancel(vc, NULL);
-
 	return NULL;
 }
 
@@ -1747,7 +1769,7 @@ static void private_chat_callback(const struct ofono_error *error, void *data)
 
 		if (vc->multiparty_list->next == NULL) {
 			g_slist_free(vc->multiparty_list);
-			vc->multiparty_list = 0;
+			vc->multiparty_list = NULL;
 		}
 	}
 
@@ -1769,7 +1791,7 @@ static DBusMessage *multiparty_private_chat(DBusConnection *conn,
 	unsigned int id;
 	GSList *l;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &callpath,
@@ -1834,7 +1856,7 @@ static void multiparty_create_callback(const struct ofono_error *error,
 	 * active calls and add them to the multiparty list
 	 */
 	old = vc->multiparty_list;
-	vc->multiparty_list = 0;
+	vc->multiparty_list = NULL;
 
 	vc->multiparty_list = g_slist_concat(vc->multiparty_list,
 						voicecalls_held_list(vc));
@@ -1867,7 +1889,7 @@ static DBusMessage *multiparty_create(DBusConnection *conn,
 {
 	struct ofono_voicecall *vc = data;
 
-	if (vc->pending)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (!voicecalls_have_held(vc) || !voicecalls_have_active(vc))
@@ -1888,7 +1910,7 @@ static DBusMessage *multiparty_hangup(DBusConnection *conn,
 {
 	struct ofono_voicecall *vc = data;
 
-	if (vc->pending || vc->release_list)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return __ofono_error_busy(msg);
 
 	if (vc->driver->release_specific == NULL)
@@ -1930,9 +1952,8 @@ static DBusMessage *multiparty_hangup(DBusConnection *conn,
 	}
 
 	/* Fall back to the old-fashioned way */
-	voicecalls_release_queue(vc, vc->multiparty_list);
-	vc->multi_release.cb = voicecalls_release_done;
-	vc->multi_release.data = vc;
+	voicecalls_release_queue(vc, vc->multiparty_list,
+					voicecalls_release_done);
 	voicecalls_release_next(vc);
 
 out:
@@ -2121,7 +2142,7 @@ void ofono_voicecall_disconnected(struct ofono_voicecall *vc, int id,
 
 			voicecall_emit_multiparty(v, FALSE);
 			g_slist_free(vc->multiparty_list);
-			vc->multiparty_list = 0;
+			vc->multiparty_list = NULL;
 		}
 	}
 
@@ -2196,6 +2217,42 @@ void ofono_voicecall_notify(struct ofono_voicecall *vc,
 		goto error;
 	}
 
+	if (vc->flags & VOICECALL_FLAG_STK_MODEM_CALLSETUP) {
+		struct dial_request *req = vc->dial_req;
+		const char *number = phone_number_to_string(&req->ph);
+
+		if (!strcmp(number, "112")) {
+			struct ofono_modem *modem =
+					__ofono_atom_get_modem(vc->atom);
+
+			__ofono_modem_inc_emergency_mode(modem);
+		}
+
+		if (v->call->clip_validity == CLIP_VALIDITY_NOT_AVAILABLE) {
+			char *number = v->call->phone_number.number;
+
+			v->call->phone_number.type = req->ph.type;
+			strncpy(number, req->ph.number,
+					OFONO_MAX_PHONE_NUMBER_LENGTH);
+			v->call->clip_validity = CLIP_VALIDITY_VALID;
+			number[OFONO_MAX_PHONE_NUMBER_LENGTH] = '\0';
+		}
+
+		v->message = req->message;
+		v->icon_id = req->icon_id;
+
+		req->message = NULL;
+		req->call = v;
+
+		/*
+		 * TS 102 223 Section 6.4.13: The terminal shall not store
+		 * in the UICC the call set-up details (called party number
+		 * and associated parameters)
+		 */
+		v->untracked = TRUE;
+		vc->flags &= ~VOICECALL_FLAG_STK_MODEM_CALLSETUP;
+	}
+
 	v->detect_time = time(NULL);
 
 	if (!voicecall_dbus_register(v)) {
@@ -2243,7 +2300,7 @@ static void multirelease_callback(const struct ofono_error *error, void *data)
 		return;
 	}
 
-	vc->multi_release.cb(error, vc->multi_release.data);
+	vc->release_queue_done_cb(error, vc);
 }
 
 static void emit_en_list_changed(struct ofono_voicecall *vc)
@@ -2465,6 +2522,14 @@ static void emulator_hfp_unregister(struct ofono_atom *atom)
 						OFONO_ATOM_TYPE_EMULATOR_HFP,
 						emulator_remove_handler,
 						"+VTS");
+	__ofono_modem_foreach_registered_atom(modem,
+						OFONO_ATOM_TYPE_EMULATOR_HFP,
+						emulator_remove_handler,
+						"D");
+	__ofono_modem_foreach_registered_atom(modem,
+						OFONO_ATOM_TYPE_EMULATOR_HFP,
+						emulator_remove_handler,
+						"+BLDN");
 
 	__ofono_modem_remove_atom_watch(modem, vc->hfp_watch);
 }
@@ -2647,10 +2712,14 @@ static void sim_state_watch(enum ofono_sim_state new_state, void *user)
 
 		free_sim_ecc_numbers(vc, FALSE);
 		set_new_ecc(vc);
+
+		voicecall_close_settings(vc);
+		break;
 	case OFONO_SIM_STATE_READY:
 		voicecall_load_settings(vc);
 		break;
-	default:
+	case OFONO_SIM_STATE_LOCKED_OUT:
+		voicecall_close_settings(vc);
 		break;
 	}
 }
@@ -2678,13 +2747,99 @@ static void sim_watch(struct ofono_atom *atom,
 
 static void emulator_generic_cb(const struct ofono_error *error, void *data)
 {
-	struct ofono_emulator *em = data;
+	struct ofono_voicecall *vc = data;
 
-	ofono_emulator_send_final(em, error);
+	if (vc->pending_em == NULL)
+		return;
+
+	ofono_emulator_send_final(vc->pending_em, error);
+	vc->pending_em = NULL;
 }
 
+static void emulator_mpty_join_cb(const struct ofono_error *error, void *data)
+{
+	struct ofono_voicecall *vc = data;
+	GSList *old;
+
+	if (vc->pending_em != NULL)
+		ofono_emulator_send_final(vc->pending_em, error);
+
+	vc->pending_em = NULL;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		return;
+
+	/*
+	 * We just created a multiparty call, gather all held
+	 * active calls and add them to the multiparty list
+	 */
+	old = vc->multiparty_list;
+	vc->multiparty_list = NULL;
+
+	vc->multiparty_list = g_slist_concat(vc->multiparty_list,
+						voicecalls_held_list(vc));
+
+	vc->multiparty_list = g_slist_concat(vc->multiparty_list,
+						voicecalls_active_list(vc));
+
+	vc->multiparty_list = g_slist_sort(vc->multiparty_list,
+						call_compare);
+
+	if (g_slist_length(vc->multiparty_list) < 2) {
+		ofono_error("Created multiparty call, but size is less than 2"
+				" panic!");
+		g_slist_free(old);
+		return;
+	}
+
+	voicecalls_multiparty_changed(old, vc->multiparty_list);
+	g_slist_free(old);
+}
+
+static void emulator_mpty_private_chat_cb(const struct ofono_error *error,
+							void *data)
+{
+	struct ofono_voicecall *vc = data;
+	GSList *old;
+	GSList *l;
+
+	if (vc->pending_em != NULL)
+		ofono_emulator_send_final(vc->pending_em, error);
+
+	vc->pending_em = NULL;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR)
+		return;
+
+	old = g_slist_copy(vc->multiparty_list);
+
+	l = g_slist_find_custom(vc->multiparty_list,
+			GINT_TO_POINTER(vc->pending_id), call_compare_by_id);
+
+	if (l) {
+		vc->multiparty_list =
+			g_slist_remove(vc->multiparty_list, l->data);
+
+		if (vc->multiparty_list->next == NULL) {
+			g_slist_free(vc->multiparty_list);
+			vc->multiparty_list = NULL;
+		}
+	}
+
+	voicecalls_multiparty_changed(old, vc->multiparty_list);
+	g_slist_free(old);
+}
+
+#define CHECK_BUSY(vc, em, result)				\
+	if (vc->pending || vc->dial_req || vc->pending_em) {	\
+		result.error = 126;				\
+		result.type = OFONO_ERROR_TYPE_CME;		\
+		ofono_emulator_send_final(em, &result);		\
+	}							\
+
 static void emulator_ata_cb(struct ofono_emulator *em,
-			struct ofono_emulator_request *req, void *userdata)
+				struct ofono_emulator_request *req,
+				void *userdata)
 {
 	struct ofono_voicecall *vc = userdata;
 	struct ofono_error result;
@@ -2693,13 +2848,16 @@ static void emulator_ata_cb(struct ofono_emulator *em,
 
 	switch (ofono_emulator_request_get_type(req)) {
 	case OFONO_EMULATOR_REQUEST_TYPE_COMMAND_ONLY:
+		CHECK_BUSY(vc, em, result)
+
 		if (!voicecalls_have_incoming(vc))
 			goto fail;
 
 		if (vc->driver->answer == NULL)
 			goto fail;
 
-		vc->driver->answer(vc, emulator_generic_cb, em);
+		vc->pending_em = em;
+		vc->driver->answer(vc, emulator_generic_cb, vc);
 		break;
 
 	default:
@@ -2721,7 +2879,10 @@ static void emulator_chup_cb(struct ofono_emulator *em,
 
 	switch (ofono_emulator_request_get_type(req)) {
 	case OFONO_EMULATOR_REQUEST_TYPE_COMMAND_ONLY:
-		if (vc->release_list)
+		if (vc->pending || vc->pending_em)
+			goto fail;
+
+		if (vc->dial_req && vc->dial_req->call == NULL)
 			goto fail;
 
 		if (vc->driver->release_specific == NULL &&
@@ -2729,7 +2890,8 @@ static void emulator_chup_cb(struct ofono_emulator *em,
 			goto fail;
 
 		if (vc->driver->hangup_active) {
-			vc->driver->hangup_active(vc, emulator_generic_cb, em);
+			vc->pending_em = em;
+			vc->driver->hangup_active(vc, emulator_generic_cb, vc);
 			goto done;
 		}
 
@@ -2747,12 +2909,11 @@ static void emulator_chup_cb(struct ofono_emulator *em,
 		if (vc->release_list == NULL)
 			goto fail;
 
-		vc->multi_release.cb = emulator_generic_cb;
-		vc->multi_release.data = em;
+		vc->pending_em = em;
+		vc->release_queue_done_cb = emulator_generic_cb;
 		voicecalls_release_next(vc);
 
 done:
-		dial_request_user_cancel(vc, NULL);
 		break;
 
 	default:
@@ -2841,6 +3002,8 @@ static void emulator_chld_cb(struct ofono_emulator *em,
 		if (!ofono_emulator_request_next_number(req, &chld))
 			goto fail;
 
+		CHECK_BUSY(vc, em, result)
+
 		switch (chld) {
 		case 0:
 			if (vc->driver->set_udub == NULL)
@@ -2849,42 +3012,52 @@ static void emulator_chld_cb(struct ofono_emulator *em,
 			if (vc->driver->release_all_held == NULL)
 				goto fail;
 
+			vc->pending_em = em;
+
 			if (voicecalls_have_waiting(vc)) {
 				vc->driver->set_udub(vc,
-						emulator_generic_cb, em);
+						emulator_generic_cb, vc);
 				return;
 			}
 
 			vc->driver->release_all_held(vc,
-					emulator_generic_cb, em);
+					emulator_generic_cb, vc);
 			return;
 		case 1:
 			if (vc->driver->release_all_active == NULL)
 				goto fail;
 
+			vc->pending_em = em;
 			vc->driver->release_all_active(vc,
-					emulator_generic_cb, em);
+					emulator_generic_cb, vc);
 			return;
 		case 2:
 			if (vc->driver->hold_all_active == NULL)
 				goto fail;
 
+			vc->pending_em = em;
 			vc->driver->hold_all_active(vc,
-					emulator_generic_cb, em);
+					emulator_generic_cb, vc);
 			return;
 		case 3:
 			if (vc->driver->create_multiparty == NULL)
 				goto fail;
 
+			if (!voicecalls_have_held(vc) ||
+					!voicecalls_have_active(vc))
+				goto fail;
+
+			vc->pending_em = em;
 			vc->driver->create_multiparty(vc,
-					emulator_generic_cb, em);
+					emulator_mpty_join_cb, vc);
 			return;
 		case 4:
 			if (vc->driver->transfer == NULL)
 				goto fail;
 
+			vc->pending_em = em;
 			vc->driver->transfer(vc,
-					emulator_generic_cb, em);
+					emulator_generic_cb, vc);
 			return;
 		default:
 			break;
@@ -2894,17 +3067,36 @@ static void emulator_chld_cb(struct ofono_emulator *em,
 			if (vc->driver->release_specific == NULL)
 				goto fail;
 
+			vc->pending_em = em;
 			vc->driver->release_specific(vc, chld - 10,
-					emulator_generic_cb, em);
+						emulator_generic_cb, vc);
 			return;
 		}
 
 		if (chld >= 21 && chld <= 27) {
+			GSList *l;
+			unsigned int id = chld - 20;
+
 			if (vc->driver->private_chat == NULL)
 				goto fail;
 
-			vc->driver->private_chat(vc, chld - 20,
-					emulator_generic_cb, em);
+			for (l = vc->multiparty_list; l; l = l->next) {
+				struct voicecall *v = l->data;
+				if (v->call->id == id)
+					break;
+			}
+
+			if (l == NULL)
+				goto fail;
+
+			if (voicecalls_have_held(vc))
+				goto fail;
+
+			vc->pending_em = em;
+			vc->pending_id = id;
+
+			vc->driver->private_chat(vc, id,
+					emulator_mpty_private_chat_cb, vc);
 			return;
 		}
 
@@ -2964,8 +3156,8 @@ static void emulator_vts_cb(struct ofono_emulator *em,
 		if (str == NULL)
 			break;
 
-		if (!g_ascii_isdigit(str[0]) && str[0] != '*' && str[0] != '#'
-				&& (str[0] < 'A' || str[0] > 'D'))
+		if (!g_ascii_isdigit(str[0]) && str[0] != '*' &&
+				str[0] != '#' && (str[0] < 'A' || str[0] > 'D'))
 			break;
 
 		if (str[1] != '\0')
@@ -2986,22 +3178,188 @@ static void emulator_vts_cb(struct ofono_emulator *em,
 	ofono_emulator_send_final(em, &result);
 }
 
+static void emulator_dial_callback(const struct ofono_error *error, void *data)
+{
+	struct ofono_voicecall *vc = data;
+	gboolean need_to_emit;
+	struct voicecall *v;
+	const char *number;
+	GError *err = NULL;
+
+	number = g_key_file_get_string(vc->settings, SETTINGS_GROUP,
+					"Number", &err);
+
+	v = dial_handle_result(vc, error, number, &need_to_emit);
+
+	if (v == NULL) {
+		struct ofono_modem *modem = __ofono_atom_get_modem(vc->atom);
+
+		if (is_emergency_number(vc, number) == TRUE)
+			__ofono_modem_dec_emergency_mode(modem);
+	}
+
+	if (vc->pending_em)
+		ofono_emulator_send_final(vc->pending_em, error);
+
+	vc->pending_em = NULL;
+
+	if (need_to_emit)
+		voicecalls_emit_call_added(vc, v);
+}
+
+static void emulator_dial(struct ofono_emulator *em, struct ofono_voicecall *vc,
+				const char *number)
+{
+	struct ofono_error result;
+	int err;
+
+	result.error = 0;
+
+	if (vc->pending || vc->dial_req || vc->pending_em) {
+		result.type = OFONO_ERROR_TYPE_FAILURE;
+		goto send;
+	}
+
+	vc->pending_em = em;
+
+	err = voicecall_dial(vc, number, OFONO_CLIR_OPTION_DEFAULT,
+					emulator_dial_callback, vc);
+
+	if (err >= 0)
+		return;
+
+	vc->pending_em = NULL;
+
+	switch (err) {
+	case -ENETDOWN:
+		result.error = 30;
+		result.type = OFONO_ERROR_TYPE_CME;
+		break;
+
+	default:
+		result.type = OFONO_ERROR_TYPE_FAILURE;
+	}
+
+send:
+	ofono_emulator_send_final(em, &result);
+}
+
+static void emulator_atd_cb(struct ofono_emulator *em,
+			struct ofono_emulator_request *req, void *userdata)
+{
+	struct ofono_voicecall *vc = userdata;
+	struct ofono_modem *modem = __ofono_atom_get_modem(vc->atom);
+	const char *str;
+	size_t len;
+	char number[OFONO_MAX_PHONE_NUMBER_LENGTH + 1];
+	struct ofono_error result;
+
+	switch (ofono_emulator_request_get_type(req)) {
+	case OFONO_EMULATOR_REQUEST_TYPE_SET:
+		str = ofono_emulator_request_get_raw(req);
+
+		if (str == NULL || str[0] == '\0')
+			goto fail;
+
+		len = strlen(str);
+
+		if (len > OFONO_MAX_PHONE_NUMBER_LENGTH + 1 ||
+				str[len - 1] != ';')
+			goto fail;
+
+		if (len == 3 && str[0] == '>' && str[1] == '1') {
+			struct ofono_atom *mw_atom;
+			struct ofono_message_waiting *mw;
+			const struct ofono_phone_number *ph;
+			const char *num;
+
+			mw_atom = __ofono_modem_find_atom(modem,
+					OFONO_ATOM_TYPE_MESSAGE_WAITING);
+
+			if (mw_atom == NULL)
+				goto fail;
+
+			mw = __ofono_atom_get_data(mw_atom);
+			ph = __ofono_message_waiting_get_mbdn(mw, 0);
+
+			if (ph == NULL)
+				goto fail;
+
+			num = phone_number_to_string(ph);
+
+			emulator_dial(em, vc, num);
+		} else {
+			strncpy(number, str, len - 1);
+			number[len - 1] = '\0';
+
+			emulator_dial(em, vc, number);
+		}
+
+		break;
+
+	default:
+fail:
+		result.error = 0;
+		result.type = OFONO_ERROR_TYPE_FAILURE;
+		ofono_emulator_send_final(em, &result);
+	};
+}
+
+static void emulator_bldn_cb(struct ofono_emulator *em,
+			struct ofono_emulator_request *req, void *userdata)
+{
+	struct ofono_voicecall *vc = userdata;
+	const char *number;
+	struct ofono_error result;
+	GError *error = NULL;
+
+	switch (ofono_emulator_request_get_type(req)) {
+	case OFONO_EMULATOR_REQUEST_TYPE_COMMAND_ONLY:
+		if (vc->settings == NULL)
+			goto fail;
+
+		number = g_key_file_get_string(vc->settings, SETTINGS_GROUP,
+						"Number", &error);
+		if (number == NULL || number[0] == '\0')
+			goto fail;
+
+		emulator_dial(em, vc, number);
+		break;
+
+	default:
+fail:
+		result.error = 0;
+		result.type = OFONO_ERROR_TYPE_FAILURE;
+		ofono_emulator_send_final(em, &result);
+	};
+}
+
 static void emulator_hfp_watch(struct ofono_atom *atom,
 				enum ofono_atom_watch_condition cond,
 				void *data)
 {
 	struct ofono_emulator *em = __ofono_atom_get_data(atom);
+	struct ofono_voicecall *vc = data;
+
+	switch (cond) {
+	case OFONO_ATOM_WATCH_CONDITION_UNREGISTERED:
+		if (vc->pending_em == em)
+			vc->pending_em = NULL;
 
-	if (cond != OFONO_ATOM_WATCH_CONDITION_REGISTERED)
 		return;
+	case OFONO_ATOM_WATCH_CONDITION_REGISTERED:
+		break;
+	}
 
-	notify_emulator_call_status(data);
+	notify_emulator_call_status(vc);
 
-	ofono_emulator_add_handler(em, "A", emulator_ata_cb, data, NULL);
-	ofono_emulator_add_handler(em, "+CHUP", emulator_chup_cb, data, NULL);
-	ofono_emulator_add_handler(em, "+CLCC", emulator_clcc_cb, data, NULL);
-	ofono_emulator_add_handler(em, "+CHLD", emulator_chld_cb, data, NULL);
-	ofono_emulator_add_handler(em, "+VTS", emulator_vts_cb, data, NULL);
+	ofono_emulator_add_handler(em, "A", emulator_ata_cb, vc, NULL);
+	ofono_emulator_add_handler(em, "+CHUP", emulator_chup_cb, vc, NULL);
+	ofono_emulator_add_handler(em, "+CLCC", emulator_clcc_cb, vc, NULL);
+	ofono_emulator_add_handler(em, "+CHLD", emulator_chld_cb, vc, NULL);
+	ofono_emulator_add_handler(em, "+VTS", emulator_vts_cb, vc, NULL);
+	ofono_emulator_add_handler(em, "D", emulator_atd_cb, vc, NULL);
+	ofono_emulator_add_handler(em, "+BLDN", emulator_bldn_cb, vc, NULL);
 }
 
 void ofono_voicecall_register(struct ofono_voicecall *vc)
@@ -3072,7 +3430,7 @@ int ofono_voicecall_get_next_callid(struct ofono_voicecall *vc)
 ofono_bool_t __ofono_voicecall_is_busy(struct ofono_voicecall *vc,
 					enum ofono_voicecall_interaction type)
 {
-	if (vc->pending || vc->dial_req)
+	if (vc->pending || vc->dial_req || vc->pending_em)
 		return TRUE;
 
 	switch (type) {
@@ -3207,7 +3565,7 @@ int __ofono_voicecall_dial(struct ofono_voicecall *vc,
 
 	/* TODO: parse the tones to dial after call connected */
 	req->ph.type = addr_type;
-	strncpy(req->ph.number, addr, 20);
+	strncpy(req->ph.number, addr, OFONO_MAX_PHONE_NUMBER_LENGTH);
 
 	vc->dial_req = req;
 
@@ -3352,6 +3710,37 @@ void __ofono_voicecall_tone_cancel(struct ofono_voicecall *vc, int id)
 	}
 }
 
+void __ofono_voicecall_set_alpha_and_icon_id(struct ofono_voicecall *vc,
+						const char *addr, int addr_type,
+						const char *message,
+						unsigned char icon_id)
+{
+	struct dial_request *req;
+
+	req = g_new0(struct dial_request, 1);
+
+	req->message = g_strdup(message);
+	req->icon_id = icon_id;
+
+	req->ph.type = addr_type;
+	strncpy(req->ph.number, addr, OFONO_MAX_PHONE_NUMBER_LENGTH);
+
+	vc->dial_req = req;
+
+	vc->flags |= VOICECALL_FLAG_STK_MODEM_CALLSETUP;
+}
+
+void __ofono_voicecall_clear_alpha_and_icon_id(struct ofono_voicecall *vc)
+{
+	g_free(vc->dial_req->message);
+	vc->dial_req->message = NULL;
+
+	g_free(vc->dial_req);
+	vc->dial_req = NULL;
+
+	vc->flags &= ~VOICECALL_FLAG_STK_MODEM_CALLSETUP;
+}
+
 static void ssn_mt_forwarded_notify(struct ofono_voicecall *vc,
 					unsigned int id, int code,
 					const struct ofono_phone_number *ph)
